from typing import Any

import serpyco

from .base import Dataclass, test_object

_serializer = serpyco.Serializer(Dataclass)

test_object = test_object


def load(data: dict[str, Any], validate: bool = True) -> Dataclass:
    return _serializer.load(data, validate=validate)


def dump(obj: Dataclass) -> dict[str, Any]:
    return _serializer.dump(obj)



from typing import Any

import marshmallow_dataclass

from .base import Dataclass, test_object

_schema = marshmallow_dataclass.class_schema(Dataclass)()

test_object = test_object


def load(data: dict[str, Any], validate: bool = True) -> Dataclass:
    return _schema.load(data)


def dump(obj: Dataclass) -> dict[str, Any]:
    return _schema.dump(obj)



import pytest

from .libs import marshmallow, pydantic, serpyco, serpyco_rs, mashumaro

serializers = {
    "serpyco": serpyco,
    "serpyco_rs": serpyco_rs,
    "pydantic": pydantic,
    "marshmallow": marshmallow,
    "mashumaro": mashumaro,
}


@pytest.mark.parametrize("lib", serializers.keys())
def test_dump(benchmark, lib):
    serializer = serializers[lib]
    serializer.dump(serializer.test_object)  # warmup

    benchmark.group = "dump"
    benchmark.extra_info["lib"] = lib
    benchmark.extra_info["correct"] = serializer.load(serializer.dump(serializer.test_object)) == serializer.test_object
    benchmark(serializer.dump, serializer.test_object)


@pytest.mark.parametrize("lib", serializers.keys())
def test_load(benchmark, lib):
    serializer = serializers[lib]
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict, validate=False)  # warmup

    benchmark.group = "load"
    benchmark.extra_info["lib"] = lib
    benchmark.extra_info["correct"] = serializer.load(serializer.dump(serializer.test_object)) == serializer.test_object
    benchmark(serializer.load, test_dict, validate=False)


@pytest.mark.parametrize("lib", serializers.keys())
def test_load_validate(benchmark, lib):
    serializer = serializers[lib]
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict, validate=True)  # warmup

    benchmark.group = "load with validate"
    benchmark.extra_info["lib"] = lib
    benchmark.extra_info["correct"] = serializer.load(serializer.dump(serializer.test_object)) == serializer.test_object

    benchmark(serializer.load, test_dict, validate=True)



from dataclasses import dataclass
from typing import Optional
from unittest.mock import ANY

from serpyco_rs import Serializer
from serpyco_rs._describe import EntityField, EntityType, OptionalType, RecursionHolder, StringType, describe_type
from serpyco_rs.metadata import NoFormat


@dataclass
class Node:
    value: str
    next: Optional["Node"] = None


@dataclass
class Root:
    head: Node


def test_describe__recursive_type__parsed():
    assert describe_type(Root) == EntityType(
        cls=Root,
        name=ANY,
        fields=[
            EntityField(
                name="head",
                dict_key="head",
                type=EntityType(
                    cls=Node,
                    name=ANY,
                    fields=[
                        EntityField(
                            name="value",
                            dict_key="value",
                            type=StringType(),
                        ),
                        EntityField(
                            name="next",
                            dict_key="next",
                            default=None,
                            type=OptionalType(
                                inner=RecursionHolder(
                                    cls=Node,
                                    name=ANY,
                                    field_format=NoFormat,
                                    state=ANY,
                                ),
                            ),
                        ),
                    ],
                    doc=ANY,
                ),
            )
        ],
        doc=ANY,
    )


def test_serializer():
    serializer = Serializer(Root)

    linked_list = Root(
        head=Node(
            value="1",
            next=Node(value="2"),
        ),
    )

    assert serializer.dump(linked_list) == {"head": {"next": {"next": None, "value": "2"}, "value": "1"}}
    assert serializer.load({"head": {"next": {"next": None, "value": "2"}, "value": "1"}}) == linked_list



from dataclasses import dataclass
from typing import Optional


@dataclass
class Nested:
    """
    A nested type for Dataclass
    """

    name: str


@dataclass
class Dataclass:
    """
    A Dataclass class
    """

    name: str
    value: int
    f: float
    b: bool
    nest: list[Nested]
    many: list[int]
    option: Optional[str] = None


def make_test_object(cls, nested_cls):
    return cls(
        name="Foo",
        value=42,
        f=12.34,
        b=True,
        nest=[nested_cls(name="Bar_{}".format(index)) for index in range(0, 1000)],
        many=[1, 2, 3],
    )


test_object = make_test_object(Dataclass, Nested)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


@dataclass
class Schema:
    type: str | None = None
    title: str | None = None
    description: str | None = None
    default: Any | None = None
    deprecated: bool | None = None
    enum: list[Any] | None = None

    allOf: list[Schema] | None = None
    anyOf: list[Schema] | None = None
    oneOf: list[Schema] | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = {
            "type": self.type,
            "title": self.title,
            "description": self.description,
            "default": self.default,
            "deprecated": self.deprecated,
            "enum": self.enum,
            "allOf": [item.dump(definitions) for item in self.allOf] if self.allOf else None,
            "anyOf": [item.dump(definitions) for item in self.anyOf] if self.anyOf else None,
            "oneOf": [item.dump(definitions) for item in self.oneOf] if self.oneOf else None,
        }
        return {k: v for k, v in data.items() if v is not None}


@dataclass
class Boolean(Schema):
    type: str = "boolean"


@dataclass
class Null(Schema):
    type: str = "null"


@dataclass
class StringType(Schema):
    type: str = "string"
    minLength: int | None = None
    maxLength: int | None = None
    pattern: str | None = None
    format: str | None = None
    # todo: enum https://json-schema.org/understanding-json-schema/reference/string.html#built-in-formats

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        data = {
            "minLength": self.minLength,
            "maxLength": self.maxLength,
            "pattern": self.pattern,
            "format": self.format,
            **data,
        }
        return {k: v for k, v in data.items() if v is not None}


@dataclass
class NumberType(Schema):
    type: str = "number"
    multipleOf: int | None = None
    minimum: float | None = None
    maximum: float | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        data = {
            "multipleOf": self.multipleOf,
            "minimum": self.minimum,
            "maximum": self.maximum,
            **data,
        }
        return {k: v for k, v in data.items() if v is not None}


@dataclass
class IntegerType(NumberType):
    type: str = "integer"


@dataclass
class ObjectType(Schema):
    name: str | None = None
    type: str = "object"
    properties: dict[str, Schema] | None = None
    additionalProperties: bool | Schema | None = None
    required: list[str] | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        data = {
            "properties": {k: v.dump(definitions) for k, v in self.properties.items()} if self.properties else None,
            "additionalProperties": (
                self.additionalProperties.dump(definitions)
                if isinstance(self.additionalProperties, Schema)
                else self.additionalProperties
            ),
            "required": self.required,
            **data,
        }
        data = {k: v for k, v in data.items() if v is not None}
        if not self.name:
            return data
        definitions[self.name] = data
        return {
            "$ref": f"#/definitions/{self.name}",
        }


@dataclass
class ArrayType(Schema):
    type: str = "array"
    items: Schema | None = None
    prefixItems: list[Schema] | None = None
    minItems: int | None = None
    maxItems: int | None = None
    uniqueItems: bool | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        data = {
            "items": self.items.dump(definitions) if self.items else None,
            "prefixItems": [i.dump(definitions) for i in self.prefixItems] if self.prefixItems else None,
            "minItems": self.minItems,
            "maxItems": self.maxItems,
            "uniqueItems": self.uniqueItems,
            **data,
        }
        return {k: v for k, v in data.items() if v is not None}


@dataclass
class RefType(Schema):
    ref: str | None = None

    def dump(self, definitions: dict[str, Any]) -> dict[str, Any]:
        data = super().dump(definitions)
        return {
            "$ref": self.ref,
            **data,
        }



import sys
from dataclasses import dataclass
from datetime import datetime, time
from decimal import Decimal
from enum import Enum
from typing import Annotated, Any, Optional
from uuid import UUID

import pytest
from serpyco_rs._describe import describe_type
from serpyco_rs._json_schema import get_json_schema
from serpyco_rs.metadata import CamelCase, Max, MaxLength, Min, MinLength


def test_to_json_schema():
    class EnumCls(Enum):
        a = "a"

    @dataclass
    class InnerData:
        """Some important entity"""

        foo: str

    @dataclass
    class Data:
        """Docs"""

        a: Annotated[int, Min(0), Max(10)]
        """A field with bounds"""
        b: float
        c: Decimal
        d: bool
        e: Annotated[str, MinLength(1), MaxLength(5)]
        f: UUID
        g: time
        h: datetime
        i: EnumCls
        j: InnerData
        k: list[int]
        l: tuple[int, str, InnerData]
        m: dict[str, int]
        n: Any
        some_filed: Annotated[str, CamelCase]

    schema = get_json_schema(describe_type(Data))

    assert schema == {
        "$ref": "#/definitions/tests._json_schema.test_convert.Data[no_format]",
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "definitions": {
            "tests._json_schema.test_convert.Data[no_format]": {
                "description": "Docs",
                "properties": {
                    "a": {
                        "description": "A " "field " "with " "bounds",
                        "maximum": 10,
                        "minimum": 0,
                        "type": "integer",
                    },
                    "b": {"type": "number"},
                    "c": {"oneOf": [{"type": "string"}, {"type": "number"}]},
                    "d": {"type": "boolean"},
                    "e": {"maxLength": 5, "minLength": 1, "type": "string"},
                    "f": {"format": "uuid", "type": "string"},
                    "g": {
                        "format": "regex",
                        "pattern": "^[0-9][0-9]:[0-9][0-9](:[0-9][0-9](\\.[0-9]+)?)??(([+-][0-9][0-9]:?[0-9][0-9])|Z)?$",
                        "type": "string",
                    },
                    "h": {
                        "format": "regex",
                        "pattern": "^[0-9]{4}-[0-9][0-9]-[0-9][0-9]T[0-9][0-9]:[0-9][0-9]:[0-9][0-9](\\.[0-9]+)?(([+-][0-9][0-9]:[0-9][0-9])|Z)?$",
                        "type": "string",
                    },
                    "i": {"enum": ["a"]},
                    "j": {"$ref": "#/definitions/tests._json_schema.test_convert.InnerData[no_format]"},
                    "k": {"items": {"type": "integer"}, "type": "array"},
                    "l": {
                        "maxItems": 3,
                        "minItems": 3,
                        "prefixItems": [
                            {"type": "integer"},
                            {"type": "string"},
                            {"$ref": "#/definitions/tests._json_schema.test_convert.InnerData[no_format]"},
                        ],
                        "type": "array",
                    },
                    "m": {"additionalProperties": {"type": "integer"}, "type": "object"},
                    "n": {},
                    "someFiled": {"type": "string"},
                },
                "required": ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "someFiled"],
                "type": "object",
            },
            "tests._json_schema.test_convert.InnerData[no_format]": {
                "description": "Some " "important " "entity",
                "properties": {"foo": {"type": "string"}},
                "required": ["foo"],
                "type": "object",
            },
        },
    }


@pytest.mark.skipif(sys.version_info < (3, 10), reason="New style unions available after 3.10")
def test_to_json_schema__new_union_syntax():
    @dataclass
    class Data:
        """Docs"""

        a: int | None

    schema = get_json_schema(describe_type(Data))

    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": "#/definitions/tests._json_schema.test_convert.Data[no_format]",
        "definitions": {
            "tests._json_schema.test_convert.Data[no_format]": {
                "description": "Docs",
                "properties": {"a": {"anyOf": [{"type": "null"}, {"type": "integer"}]}},
                "required": ["a"],
                "type": "object",
            }
        },
    }


@dataclass
class TreeNode:
    """Node"""

    data: str
    left: Optional["TreeNode"] = None
    right: Optional["TreeNode"] = None


def test_to_json_schema__recursive_type():

    schema = get_json_schema(describe_type(TreeNode))

    assert schema == {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        "$ref": "#/definitions/tests._json_schema.test_convert.TreeNode[no_format]",
        "definitions": {
            "tests._json_schema.test_convert.TreeNode[no_format]": {
                "description": "Node",
                "properties": {
                    "data": {"type": "string"},
                    "left": {
                        "anyOf": [
                            {"type": "null"},
                            {"$ref": "#/definitions/tests._json_schema.test_convert.TreeNode[no_format]"},
                        ]
                    },
                    "right": {
                        "anyOf": [
                            {"type": "null"},
                            {"$ref": "#/definitions/tests._json_schema.test_convert.TreeNode[no_format]"},
                        ]
                    },
                },
                "required": ["data"],
                "type": "object",
            }
        },
    }



from functools import singledispatch
from typing import Any, Optional

from .. import _describe as describe
from ._entities import ArrayType, Boolean, IntegerType, Null, NumberType, ObjectType, RefType, Schema, StringType


def get_json_schema(t: describe.Type) -> dict[str, Any]:
    schema = to_json_schema(t)
    definitions: dict[str, Any] = {}
    schema_def = schema.dump(definitions)
    return {
        "$schema": "https://json-schema.org/draft/2020-12/schema",
        **schema_def,
        "definitions": definitions,
    }


@singledispatch
def to_json_schema(_: Any, doc: Optional[str] = None) -> Schema:
    return Schema(description=doc)


@to_json_schema.register
def _(arg: describe.StringType, doc: Optional[str] = None) -> Schema:
    return StringType(
        minLength=arg.min_length,
        maxLength=arg.max_length,
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.IntegerType, doc: Optional[str] = None) -> Schema:
    return IntegerType(
        minimum=arg.min,
        maximum=arg.max,
        description=doc,
    )


@to_json_schema.register
def _(_: describe.BytesType, doc: Optional[str] = None) -> Schema:
    return StringType(
        format="binary",
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.FloatType, doc: Optional[str] = None) -> Schema:
    return NumberType(
        minimum=arg.min,
        maximum=arg.max,
        description=doc,
    )


@to_json_schema.register
def _(_: describe.DecimalType, doc: Optional[str] = None) -> Schema:
    return Schema(
        oneOf=[
            StringType(),
            NumberType(),
        ],
        description=doc,
    )


@to_json_schema.register
def _(_: describe.BooleanType, doc: Optional[str] = None) -> Schema:
    return Boolean()


@to_json_schema.register
def _(_: describe.UUIDType, doc: Optional[str] = None) -> Schema:
    return StringType(
        format="uuid",
        description=doc,
    )


@to_json_schema.register
def _(_: describe.TimeType, doc: Optional[str] = None) -> Schema:
    iso8601_pattern = (
        r"^[0-9][0-9]:[0-9][0-9](:[0-9][0-9](\.[0-9]+)?)?"  # HH:mm:ss.ssss
        r"?(([+-][0-9][0-9]:?[0-9][0-9])|Z)?$"  # timezone
    )
    return StringType(
        format="regex",
        pattern=iso8601_pattern,
        description=doc,
    )


@to_json_schema.register
def _(_: describe.DateTimeType, doc: Optional[str] = None) -> Schema:
    iso8601_pattern = (
        r"^[0-9]{4}-[0-9][0-9]-[0-9][0-9]T"  # YYYY-MM-DD
        r"[0-9][0-9]:[0-9][0-9]:[0-9][0-9](\.[0-9]+)"  # HH:mm:ss.ssss
        r"?(([+-][0-9][0-9]:[0-9][0-9])|Z)?$"  # timezone
    )
    return StringType(
        format="regex",
        pattern=iso8601_pattern,
        description=doc,
    )


@to_json_schema.register
def _(_: describe.DateType, doc: Optional[str] = None) -> Schema:
    iso8601_pattern = r"^[0-9]{4}-[0-9][0-9]-[0-9][0-9]$"  # YYYY-MM-DD
    return StringType(
        format="regex",
        pattern=iso8601_pattern,
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.EnumType, doc: Optional[str] = None) -> Schema:
    return Schema(
        enum=[item.value for item in arg.cls],
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.OptionalType, doc: Optional[str] = None) -> Schema:
    return Schema(
        anyOf=[
            Null(),
            to_json_schema(arg.inner),
        ],
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.EntityType, doc: Optional[str] = None) -> Schema:
    return ObjectType(
        properties={prop.dict_key: to_json_schema(prop.type, prop.doc) for prop in arg.fields if not prop.is_property},
        required=[
            prop.dict_key
            for prop in arg.fields
            if not (prop.is_property or prop.default != describe.NOT_SET or prop.default_factory != describe.NOT_SET)
        ]
        or None,
        name=arg.name,
        description=arg.doc,
    )


@to_json_schema.register
def _(arg: describe.ArrayType, doc: Optional[str] = None) -> Schema:
    return ArrayType(
        items=to_json_schema(arg.item_type),
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.DictionaryType, doc: Optional[str] = None) -> Schema:
    return ObjectType(
        additionalProperties=to_json_schema(arg.value_type),
        description=doc,
    )


@to_json_schema.register
def _(arg: describe.TupleType, doc: Optional[str] = None) -> Schema:
    return ArrayType(
        prefixItems=[to_json_schema(item) for item in arg.item_types],
        minItems=len(arg.item_types),
        maxItems=len(arg.item_types),
        description=doc,
    )


@to_json_schema.register
def _(_: describe.AnyType, doc: Optional[str] = None) -> Schema:
    return Schema(description=doc)


@to_json_schema.register
def _(holder: describe.RecursionHolder, doc: Optional[str] = None) -> Schema:
    return RefType(description=doc, ref=f"#/definitions/{holder.name}")



from typing import Any, Generic, TypeVar

from ._describe import Type

_T = TypeVar("_T")

class ValidationError(Exception):
    pass

class Serializer(Generic[_T]):
    def __init__(self, py_class: Type):
        pass
    def dump(self, value: _T) -> Any:
        pass
    def load(self, data: Any) -> _T:
        pass



from typing import Annotated, Any, Generic, TypeVar, cast

from ._describe import describe_type
from ._impl import Serializer as _Serializer
from ._json_schema import JsonschemaRSValidator, Validator, get_json_schema
from .metadata import CamelCase

_T = TypeVar("_T", bound=Any)


class Serializer(Generic[_T]):
    def __init__(
        self,
        t: type[_T],
        camelcase_fields: bool = False,
        validator_cls: type[Validator] = JsonschemaRSValidator,
    ) -> None:
        if camelcase_fields:
            t = cast(type[_T], Annotated[t, CamelCase])
        type_info = describe_type(t)
        self._encoder: _Serializer[_T] = _Serializer(type_info)
        self._schema = get_json_schema(type_info)
        self._validator = validator_cls(self._schema)

    def dump(self, value: _T) -> Any:
        return self._encoder.dump(value)

    def load(self, data: Any, validate: bool = True) -> _T:
        if validate:
            self._validator.validate(data)
        return self._encoder.load(data)



import dataclasses
import sys
from collections.abc import Callable, Iterable, Mapping, Sequence
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum, IntEnum
from typing import Annotated, Any, Optional, TypeVar, Union, cast, get_origin, get_type_hints, overload
from uuid import UUID

from attributes_doc import get_attributes_doc
from typing_extensions import assert_never

from ._utils import to_camelcase
from .metadata import FiledFormat, Format, Max, MaxLength, Min, MinLength, NoFormat, Places

if sys.version_info >= (3, 10):  # pragma: no cover
    from types import UnionType
else:  # pragma: no cover
    UnionType = None

try:
    import attr
except ImportError:  # pragma: no cover
    attr = None  # type: ignore


_NoneType = type(None)

_T = TypeVar("_T")


class NotSet:
    def __repr__(self) -> str:
        return "NOT_SET"


NOT_SET = NotSet()


@dataclasses.dataclass
class Type:
    pass


@dataclasses.dataclass
class IntegerType(Type):
    min: Optional[int] = None
    max: Optional[int] = None


@dataclasses.dataclass
class StringType(Type):
    min_length: Optional[int] = None
    max_length: Optional[int] = None


@dataclasses.dataclass
class BytesType(Type):
    pass


@dataclasses.dataclass
class FloatType(Type):
    min: Optional[float] = None
    max: Optional[float] = None


@dataclasses.dataclass
class DecimalType(Type):
    places: Optional[int] = None
    min: Optional[Decimal] = None
    max: Optional[Decimal] = None


@dataclasses.dataclass
class BooleanType(Type):
    pass


@dataclasses.dataclass
class UUIDType(Type):
    pass


@dataclasses.dataclass
class TimeType(Type):
    pass


@dataclasses.dataclass
class DateTimeType(Type):
    pass


@dataclasses.dataclass
class DateType(Type):
    pass


@dataclasses.dataclass
class EnumType(Type):
    cls: type[Union[Enum, IntEnum]]


@dataclasses.dataclass
class EntityField:
    name: str
    dict_key: str
    type: Type
    doc: Optional[str] = None
    default: Any = NOT_SET
    default_factory: Union[Callable[[], Any], NotSet] = NOT_SET
    is_property: bool = False


@dataclasses.dataclass
class EntityType(Type):
    cls: type[Any]
    name: str
    fields: Sequence[EntityField]
    generics: Mapping[TypeVar, Any] = dataclasses.field(default_factory=dict)
    doc: Optional[str] = None


@dataclasses.dataclass
class OptionalType(Type):
    inner: Type


@dataclasses.dataclass
class ArrayType(Type):
    item_type: Type
    is_sequence: bool


@dataclasses.dataclass
class DictionaryType(Type):
    key_type: Type
    value_type: Type
    is_mapping: bool


@dataclasses.dataclass
class TupleType(Type):
    item_types: Sequence[Type]


@dataclasses.dataclass
class AnyType(Type):
    pass


@dataclasses.dataclass
class RecursionHolder(Type):
    cls: Any
    name: str
    field_format: FiledFormat
    state: dict[tuple[type, FiledFormat], Optional[Type]]

    def get_type(self) -> Type:
        if type_ := self.state[(self.cls, self.field_format)]:
            return type_
        raise RuntimeError("Recursive type not resolved")


def describe_type(t: Any, state: Optional[dict[tuple[type, FiledFormat], Optional[Type]]] = None) -> Type:
    state = state or {}
    parameters: tuple[Any, ...] = ()
    args: tuple[Any, ...] = ()
    metadata = _get_annotated_metadata(t)
    if get_origin(t) == Annotated:  # unwrap annotated
        t = t.__origin__
    if hasattr(t, "__origin__"):
        parameters = getattr(t.__origin__, "__parameters__", ())
        args = t.__args__
        t = t.__origin__
    # UnionType has no __origin__
    elif UnionType and isinstance(t, UnionType):  # type: ignore[truthy-function]
        args = t.__args__
        t = Union
    elif hasattr(t, "__parameters__"):
        # Если передан generic-класс без type-параметров, значит по PEP-484 заменяем все параметры на Any
        parameters = t.__parameters__
        args = (Any,) * len(parameters)

    generics = dict(zip(parameters, args))
    filed_format = _find_metadata(metadata, FiledFormat, NoFormat)
    annotation_wrapper = _wrap_annotated([filed_format])

    if (t, filed_format) in state:
        return RecursionHolder(cls=t, name=_generate_name(t, filed_format), field_format=filed_format, state=state)

    if t is Any:
        return AnyType()

    if isinstance(t, type):
        simple_type_mapping: Mapping[type, type[Type]] = {
            bytes: BytesType,
            bool: BooleanType,
            date: DateType,
            time: TimeType,
            datetime: DateTimeType,
            UUID: UUIDType,
        }

        if simple := simple_type_mapping.get(t):
            return simple()

        number_type_mapping: Mapping[type, type[IntegerType] | type[FloatType]] = {
            int: IntegerType,
            float: FloatType,
        }

        if number_type := number_type_mapping.get(t):
            min_meta = _find_metadata(metadata, Min)
            max_meta = _find_metadata(metadata, Max)
            return number_type(
                min=cast(Any, min_meta.value) if min_meta else None,
                max=cast(Any, max_meta.value) if max_meta else None,
            )

        if t is Decimal:
            min_meta = _find_metadata(metadata, Min)
            max_meta = _find_metadata(metadata, Max)
            places_meta = _find_metadata(metadata, Places)
            return DecimalType(
                min=cast(Decimal, min_meta.value) if min_meta else None,
                max=cast(Decimal, max_meta.value) if max_meta else None,
                places=places_meta.value if places_meta else None,
            )

        if t is str:
            min_length_meta = _find_metadata(metadata, MinLength)
            max_length_meta = _find_metadata(metadata, MaxLength)
            return StringType(
                min_length=min_length_meta.value if min_length_meta else None,
                max_length=max_length_meta.value if max_length_meta else None,
            )

        if t in {Sequence, list}:
            return ArrayType(
                item_type=(describe_type(annotation_wrapper(args[0]), state) if args else AnyType()),
                is_sequence=t is Sequence,
            )

        if t in {Mapping, dict}:
            return DictionaryType(
                key_type=(describe_type(annotation_wrapper(args[0]), state) if args else AnyType()),
                value_type=(describe_type(annotation_wrapper(args[1]), state) if args else AnyType()),
                is_mapping=t is Mapping,
            )

        if t is tuple:
            if not args or Ellipsis in args:
                raise RuntimeError("Variable length tuples are not supported")
            return TupleType(item_types=[describe_type(annotation_wrapper(arg), state) for arg in args])

        if issubclass(t, (Enum, IntEnum)):
            return EnumType(cls=t)

        if dataclasses.is_dataclass(t):
            state[(t, filed_format)] = None
            entity_type = _describe_dataclass(t, generics, filed_format, state)
            state[(t, filed_format)] = entity_type
            return entity_type

        if attr and attr.has(t):
            state[(t, filed_format)] = None
            entity_type = _describe_attrs(t, generics, filed_format, state)
            state[(t, filed_format)] = entity_type
            return entity_type

    if t in {Union}:
        if len(args) != 2 or _NoneType not in args:
            raise RuntimeError(f"Only Unions of one type with None are supported: {t}, {args}")
        inner = args[1] if args[0] is _NoneType else args[0]
        return OptionalType(describe_type(annotation_wrapper(inner), state))

    if isinstance(t, TypeVar):
        raise RuntimeError(f"Unfilled TypeVar: {t}")

    raise RuntimeError(f"Unknown type {t!r}")


def _describe_dataclass(
    t: type[Any],
    generics: Mapping[TypeVar, Any],
    cls_filed_format: FiledFormat,
    state: dict[tuple[type, FiledFormat], Optional[Type]],
) -> EntityType:
    docs = get_attributes_doc(t)
    try:
        types = get_type_hints(t, include_extras=True)
    except Exception:  # pylint: disable=broad-except
        types = {}

    fields = []
    for field in dataclasses.fields(t):

        type_ = _replace_generics(types.get(field.name, field.type), generics)
        if cls_filed_format:
            type_ = Annotated[type_, cls_filed_format]

        metadata = _get_annotated_metadata(type_)
        field_type = describe_type(type_, state)
        field_format = _find_metadata(metadata, FiledFormat)

        fields.append(
            EntityField(
                name=field.name,
                dict_key=_apply_format(field_format, field.name),
                doc=docs.get(field.name),
                type=field_type,
                default=(field.default if field.default is not dataclasses.MISSING else NOT_SET),
                default_factory=(
                    field.default_factory if field.default_factory is not dataclasses.MISSING else NOT_SET
                ),
                is_property=False,
            )
        )

    return EntityType(cls=t, name=_generate_name(t, cls_filed_format), fields=fields, generics=generics, doc=t.__doc__)


def _describe_attrs(
    t: type[Any],
    generics: Mapping[TypeVar, Any],
    cls_filed_format: FiledFormat,
    state: dict[tuple[type, FiledFormat], Optional[Type]],
) -> EntityType:
    assert attr is not None
    docs = get_attributes_doc(t)
    try:
        types = get_type_hints(t, include_extras=True)
    except Exception:  # pylint: disable=broad-except
        types = {}
    fields = []
    for field in attr.fields(t):  # pyright: ignore
        default = NOT_SET
        if field.default is not attr.NOTHING and not isinstance(field.default, attr.Factory):  # type: ignore[arg-type]
            default = field.default

        default_factory = NOT_SET
        if isinstance(field.default, attr.Factory):  # type: ignore[arg-type]
            default_factory = field.default.factory

        type_ = _replace_generics(types.get(field.name, field.type), generics)
        if cls_filed_format:
            type_ = Annotated[type_, cls_filed_format]

        metadata = _get_annotated_metadata(type_)
        field_type = describe_type(type_, state)
        field_format = _find_metadata(metadata, FiledFormat)

        fields.append(
            EntityField(
                name=field.name,
                dict_key=_apply_format(field_format, field.name),
                doc=docs.get(field.name),
                type=field_type,
                default=default,
                default_factory=default_factory,
                is_property=False,
            )
        )
    return EntityType(cls=t, name=_generate_name(t, cls_filed_format), fields=fields, generics=generics)


def _replace_generics(t: Any, generics: Mapping[TypeVar, Any]) -> Any:
    try:
        if parameters := getattr(t, "__parameters__", None):
            t = t[tuple(generics[parameter] for parameter in parameters)]
        if isinstance(t, TypeVar):
            t = generics[t]
    except KeyError as exc:
        raise RuntimeError(f"Unfilled TypeVar: {exc.args[0]}") from exc
    return t


@overload
def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: _T) -> _T:
    ...


@overload
def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: None = None) -> Optional[_T]:
    ...


def _find_metadata(annotations: Iterable[Any], type_: type[_T], default: Optional[_T] = None) -> Optional[_T]:
    return next((ann for ann in annotations if isinstance(ann, type_)), default)


def _wrap_annotated(annotations: Iterable[Any]) -> Callable[[_T], _T]:
    def inner(type_: _T) -> _T:
        for ann in annotations:
            type_ = Annotated[type_, ann]  # type: ignore
        return type_

    return inner


def _get_annotated_metadata(t: Any) -> tuple[Any, ...]:
    if get_origin(t) == Annotated:
        return getattr(t, "__metadata__", ())
    return ()


def _apply_format(f: Optional[FiledFormat], value: str) -> str:
    if not f or f.format is Format.no_format:
        return value
    if f.format is Format.camel_case:
        return to_camelcase(value)
    assert_never(f.format)


def _generate_name(cls: Any, field_format: FiledFormat) -> str:
    return f"{cls.__module__}.{cls.__name__}[{field_format.format.value}]"



import os
import platform

import nox

nox.options.sessions = ["test", "lint", "type_check"]
nox.options.reuse_existing_virtualenvs = True


def build(session):
    if os.environ.get("CI", None):
        # Install form wheels
        session.install("--no-index", "--no-deps", "--find-links", "wheels/", "serpyco-rs")
        session.install("--find-links", "wheels/", "serpyco-rs")
        return

    if platform.machine() == "arm64" and platform.system() == "Darwin":
        # https://github.com/Stranger6667/jsonschema-rs/issues/409
        session.install("jsonschema_rs", "--no-binary", ":all:")
    session.run_always("maturin", "develop", "-r")


@nox.session
def test(session):
    build(session)
    session.install("-r", "requirements/dev.txt")
    session.run("pytest", "-vss", "tests/")


@nox.session
def lint(session):
    build(session)
    session.install("-r", "requirements/lint.txt")

    session.cd("python/serpyco_rs")
    session.run("black", "--check", "--diff", ".")
    session.run("isort", "--check", "--diff", ".")
    session.run("ruff", ".")


@nox.session
def type_check(session):
    build(session)
    session.install("-r", "requirements/type_check.txt")

    session.cd("python/serpyco_rs")
    session.run("pyright")
    session.run("pyright", "--verifytypes", "serpyco_rs")
    session.run("mypy", ".", "--strict", "--implicit-reexport", "--pretty")


@nox.session
def bench(session):
    build(session)
    session.install("-r", "requirements/bench.txt")

    session.run(
        "pytest",
        "--verbose",
        "--benchmark-min-time=1",
        "--benchmark-max-time=5",
        "--benchmark-disable-gc",
        "--benchmark-autosave",
        "--benchmark-save-data",
        "bench",
    )



import collections
import io
import json
import os

from tabulate import tabulate

LIBRARIES = (
    "serpyco_rs",
    "serpyco",
    "mashumaro",
    "pydantic",
    "marshmallow",
)


def aggregate():
    benchmarks_dir = os.path.join(".benchmarks", os.listdir(".benchmarks")[0])
    res = collections.defaultdict(dict)
    for filename in os.listdir(benchmarks_dir):
        with open(os.path.join(benchmarks_dir, filename), "r") as fileh:
            data = json.loads(fileh.read())

        for each in data["benchmarks"]:
            res[each["group"]][each["extra_info"]["lib"]] = {
                "data": [val * 1000 for val in each["stats"]["data"]],
                "median": each["stats"]["median"] * 1000,
                "ops": each["stats"]["ops"],
                "correct": each["extra_info"]["correct"],
            }
    return res


def tab(obj):
    buf = io.StringIO()
    headers = (
        "Library",
        "Median latency (milliseconds)",
        "Operations per second",
        "Relative (latency)",
    )
    for group, val in sorted(obj.items(), reverse=True):
        buf.write("\n" + "#### " + group + "\n\n")
        table = []
        for lib in LIBRARIES:
            correct = val[lib]["correct"]
            table.append(
                [
                    lib,
                    val[lib]["median"] if correct else None,
                    "%.1f" % val[lib]["ops"] if correct else None,
                    0,
                ]
            )
        baseline = table[0][1]
        for each in table:
            each[3] = "%.2f" % (each[1] / baseline) if isinstance(each[1], float) else None
            each[1] = "%.2f" % each[1] if isinstance(each[1], float) else None
        buf.write(tabulate(table, headers, tablefmt="github") + "\n")

    print(buf.getvalue())


tab(aggregate())



from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from typing import Union


@dataclass(frozen=True)
class Min:
    value: Union[int, float, Decimal]


@dataclass(frozen=True)
class Max:
    value: Union[int, float, Decimal]


@dataclass(frozen=True)
class MinLength:
    value: int


@dataclass(frozen=True)
class MaxLength:
    value: int


@dataclass(frozen=True)
class Places:
    value: int


class Format(Enum):
    no_format = "no_format"
    camel_case = "camel_case"


@dataclass(frozen=True)
class FiledFormat:
    format: Format


CamelCase: FiledFormat = FiledFormat(Format.camel_case)
NoFormat: FiledFormat = FiledFormat(Format.no_format)
